
#include "sim.h"

/*
 * Type 1: ALU/MAC with Data and Program Memory Read
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  1  1 -PD-- -DD-- ----AMF------- -Yop- -Xop---- -PMI- -PMM- -DMI- -DMM-
 *
 */
unsigned int op_pmdmr(unsigned int opcode)
{
  fprintf(stderr, "op_pmdmr: unimplemented\n");
  return 0;
}

/*
 * Type 2: Data Memory Write (Immediate Data)
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  1  0  1 -G ------------------------ADDR------------------- --I-- --M--
 *
 */
unsigned int op_dmwim(unsigned int opcode)
{
  if (!setdmindirect((opcode>>20)&0x1, 
		     (opcode>>2)&0x3, 
		     opcode&0x3, 
		     (opcode>>4)&0x1ffff))
    fprintf(stderr,"decode: indirect data memory set failed\n");
  return 0;
}

/*
 * Type 3: Read/Write Data Memory (Immediate Address)
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  1  0  0 -D ---RGP-- ---------------ADDR------------------- ----REG----
 *
 */
unsigned int op_dmrwim(unsigned int opcode)
{
  if (opcode & BIT20) /* 1001 */
    {
      /* dm(addr) = reg */
      setdmfromreg((opcode >> 4) & 0x3fff,   /* address */
		   getregptr_rgp(opcode & 0xf, /* reg */
				 (opcode >> 18) & 0x3)); /* rgp */
    }
  else /* BIT20 off */
    {
      /* reg = dm(addr) */
      setregfromdm((opcode>>4)&0x3fff, /* addr */
		   getregptr_rgp(opcode & 0xf, /* reg */
				 (opcode>>18)&0x3)); /* rgp */
      
      /* set the internal _B register if its an I reg */
      if ( ((opcode>>18&0x3)>0) && ((opcode>>18)&0x3)<3)
	{
	  if ((opcode&0xf)<4)
	    setdagregrgp((opcode&0xf), ((opcode>>18)&0x3)+3, (opcode>>4)&0x3fff);
	}
    }
  return 0;
}

/*
 * Type 4: ALU/MAC with Data Memory Read/Write
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  1  1  G  D  Z  -----AMF----  -Yop- --Xop---  ---DREG--- --I-- --M--
 *
 * ALU/MAC Optional.
 *
 */
unsigned int op_dmrw(unsigned int opcode)
{
  /* dm(i, m) = dreg */
  unsigned int dag = (opcode & BIT20)?1:0;
  if ((opcode & BIT19) && !(opcode & BIT18))
    {
      unsigned int amf = (opcode >> 13) & 0x1f;
      unsigned int i = (opcode >> 2) & 0x3;
      unsigned int m = opcode & 0x3;
      unsigned int dreg = (opcode >> 4) & 0xf;
      setdmfromreg_indirect(dag, i, m, getregptr_rgp(dreg,0));
      if (amf) fprintf(stderr, "decode: AMF unhandled\n");
#if 0
      fprintf(stderr, "decode: unknown: dm(i%d, m%d) = %d  %s\n", 
	      dag?i+4:i, 
	      dag?m+4:m, 
	      /*regs[dreg][0], */ dreg,
	      amf?"(AMF unhandled)":"");
#endif
    }
  else /* no amf */
    {
      unsigned int data = (opcode >> 4) & 0xff;
      unsigned int i = (opcode >> 1) & 0x1;
      unsigned int m = opcode & 0x1;
      fprintf(stderr, "decode: unknown: dm(i%d, m%d) = 0x%02x\n", i, m, data); 
    }
  return 0;
}

/*
 * Type 5: ALU/MAC with Program Memory Read/Write
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  1  0  1 -D -Z -----AMF------ -Yop- --Xop--- ----DREG--- --I-- --M--
 *
 * ALU/MAC optional.
 *
 */
unsigned int op_pmrw(unsigned int opcode)
{
  if (opcode & BIT19) /* 0101 1 */
    {
      if (opcode & BIT18) /* 0101 11 */
	{
	  fprintf(stderr, "decode: unknown: 0101 11\n");
	}
      else /* 0101 10 */
	{
#if 0
	  unsigned int amf = (opcode >> 13) & 0x1f;
	  unsigned int dreg = (opcode >> 4) & 0xf;
	  unsigned int i = (opcode >> 1) & 0x1;
	  unsigned int m = opcode & 0x1;
#endif
	  fprintf(stderr, "decode: unknown: 0101 10\n");
	}
    }
  else /* 01010 */
    {
      fprintf(stderr, "decode: unknown: 0101 0\n");
    }
  return 0;
}

/*
 * Type 6: Load Data Register Immediate
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  1  0  0 -----------------------DATA-------------------- ----DREG---
 *
 */
unsigned int op_loadregim(unsigned int opcode)
{
  setregrgp(opcode & 0xf, 0, (opcode>>4)&0x3fff);
  return 0;
}

/*
 * Type 7: Load Non-Data Register Immediate
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  0  1  1 -RGP- -----------------DATA-------------------- ----REG----
 *
 */
unsigned int op_loadnondataregim(unsigned int opcode)
{
  setregrgp(opcode & 0xf, /* reg */
	    (opcode>>18) & 0x3, /* rgp */
	    (opcode>>4) & 0x3fff /* addr */);
  
  /* set the internal _B register if its an I reg */
  if ( ((opcode>>18&0x3)>0) && ((opcode>>18)&0x3)<3)
    {
      if ((opcode&0xf)<4)
	setdagregrgp((opcode&0xf), ((opcode>>18)&0x3)+3, (opcode>>4)&0x3fff);
    }
  return 0;
}

/*
 * Type 8a: ALU/MAC with Internal Data Register Move 
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  0  1  0  1 -Z -----AMF------ -Yop- ---Xop--  -DestDREG-  --SrcDREG-
 *
 *
 * Type 8b: Generate ALU Status (none = <ALU>)
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  0  1  0  1  0 -----AMF------ -Yop- ---Xop--  1  0  1  0  1  0  1  0
 *                      |
 *                       -- ALU codes only (no MAC) 
 *
 */
unsigned int op_alumove(unsigned int opcode)
{
  fprintf(stderr, "op_alumove: unimplemented\n");
  return 0;
}

/*
 * Type 9's... Conditional ALU/MAC
 *
 * Type 9a: xop * yop
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  0  1  0  0 -Z -----AMF------ -Yop- ---Xop--  0  0  0  0 ---COND----
 *
 * Type 9b: xop * xop
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  0  1  0  0 -Z -----AMF------  0  0 ---Xop--  0  0  0  1 ---COND----
 *
 * Type 9c: xop {AND,OR,XOR} constant
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  0  1  0  0 -Z -----AMF------ -YY-- ---Xop-- -CC-- -BO-- ---COND----
 *
 * Type 9d: PASS constant (constant != 0,1,-1)
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  0  1  0  0 -Z -----AMF------ -YY-- ---Xop-- -CC-- -BO-- ---COND----
 *
 */
unsigned int op_alu(unsigned int opcode)
{
  fprintf(stderr, "op_alu: unimplemented\n");
  return 0;
}

/*
 * Type 10: Conditional Jump (Immediate Addres)
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  0  0  1  1 -S ------------------ADDR------------------- ----COND---
 *
 */
unsigned int op_jump(unsigned int opcode)
{	      
  if (opcode & BIT18) /* 000111 */
    {			  
      if (testcond(opcode&0xf)) /* call */
	{
	  pc_push();
	  return (opcode >> 4) & 0x3fff;
	}
    }
  else /* 000110 */
    {
      if (testcond(opcode&0xf)) /* jump */
	return (opcode >> 4) & 0x3fff;
    }
  return 0;
}

/*
 * Type 11: Do Until
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  0  0  1  0  1 --------------------ADDR----------------- ----COND---
 *
 */
unsigned int op_dountil(unsigned int opcode)
{	      
  unsigned short addr,cond;
  addr = (opcode>>4) & 0x3fff;
  cond = opcode & 0xf;
  curdsp.sequencer.pc++;
  pc_push();
  loop_push(addr, cond);
  curdsp.sequencer.pc--;
  return 0;
}

/*
 * Type 12: Shift with Data Memory Read/Write
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  0  0  1  0  0  1 -G -D ----SF----- --Xop--- ---Dreg---- --I-- --M--
 *
 */
unsigned int op_shiftdmrw(unsigned int opcode)
{
  fprintf(stderr, "op_shiftdmrw: unimplmeneted\n");
  return 0;
}

/*
 * Type 13: Shift with Program Memory Read/Write
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  0  0  1  0  0  0  1 -D ----SF----- --Xop--- ---Dreg---- --I-- --M--
 *
 */
unsigned int op_shiftpmrw(unsigned int opcode)
{
  fprintf(stderr, "op_shiftpmrw: unimplmeneted\n");
  return 0;
}

/*
 * Type 14: Shift with Internal Data Register Move
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  0  0  1  0  0  0  0  0 ----SF----- --Xop--- --DestDREG- --SrcDREG--
 *
 */
unsigned int op_shiftmove(unsigned int opcode)
{
  fprintf(stderr, "op_shiftmove: unimplmeneted\n");
  return 0;
}

/*
 * Type 15: Shift Immediate
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  0  0  0  1  1  1  1  0 ----SF----- --XOP--- --------exponent-------
 *
 */
unsigned int op_shiftim(unsigned int opcode)
{
  fprintf(stderr, "op_shiftim: unimplmeneted\n");
  return 0;
}

/*
 * Type 16: Conditional Shift
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  0  0  0  1  1  1  0  0 ----SF----- ---XOP--  0  0  0  0 ----COND---
 *
 */
unsigned int op_shift(unsigned int opcode)
{
  fprintf(stderr, "op_shift: unimplmeneted\n");
  return 0;
}

/*
 * Type 17: Internal Data Move
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  0  0  0  1  1  0  1  0  0  0  0 -DRGP -SRGP --DestReg-- ---SrcReg--
 *
 */
unsigned int op_datamove(unsigned int opcode)
{	
  regmove((opcode>>4)&0xf, (opcode>>10)&0x3, opcode&0xf, (opcode>>8)&0x3);
  return 0;
}

/*
 * Type 18: Mode Control
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  0  0  0  1  1  0  0 -TI-- -MM-- -AS-- -OL-- -BR-- -SR-- -GM--  0  0
 *
 */
unsigned int op_modectl(unsigned int opcode)
{	
  if (((opcode>>2)&03)>=2)
    {
      if (((opcode>>2)&0x3)==2)
	set_mstatclearbit(MSTAT_GMODE);
      else
	set_mstatbit(MSTAT_GMODE);
    }
  if (((opcode>>4)&03)>=2)
    {
      if (((opcode>>4)&0x3)==2)
	set_mstatclearbit(MSTAT_SECREG);
      else
	set_mstatbit(MSTAT_SECREG);
    }
  
  if (((opcode>>6)&03)>=2)
    {
      if (((opcode>>6)&0x3)==2)
	set_mstatclearbit(MSTAT_BITREV);
      else
	set_mstatbit(MSTAT_BITREV);
    }
  
  if (((opcode>>8)&03)>=2)
    {
      if (((opcode>>8)&0x3)==2)
	set_mstatclearbit(MSTAT_AVLATCH);
      else
	set_mstatbit(MSTAT_AVLATCH);
    }
  
  if (((opcode>>0x0a)&03)>=2)
    {
      if (((opcode>>0x0a)&0x3)==2)
	set_mstatclearbit(MSTAT_ARSAT);
      else
	set_mstatbit(MSTAT_ARSAT);
    }
  
  if (((opcode>>0x0c)&03)>=2)
    {
      if (((opcode>>0x0c)&0x3)==2)
	set_mstatclearbit(MSTAT_MMODE);
      else
	set_mstatbit(MSTAT_MMODE);
    }
  
  if (((opcode>>0x0e)&03)>=2)
    {
      if (((opcode>>0x0e)&0x3)==2)
	set_mstatclearbit(MSTAT_TIMER);
      else
	set_mstatbit(MSTAT_TIMER);
    }
  return 0;
}

/*
 * Type 19: Conditional Jump (Indirect Address)
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  0  0  0  1  0  1  1  0  0  0  0  0  0  0  0 --I--  0 -S ----COND---
 *
 */
unsigned int op_jumpindirect(unsigned int opcode)
{
  fprintf(stderr, "op_jumpindirect: unimplemented\n");
  return 0;
}

/*
 * Type 20: Conditional Return 
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  0  0  0  1  0  1  0  0  0  0  0  0  0  0  0  0  0  0 -T ----COND---
 *
 */
unsigned int op_return(unsigned int opcode)
{
  if ((opcode >> 4) == 0xa001) /* rti */
    {
      if (testcond(opcode&0xf))
	seq_rti();
    }
  else if ((opcode >> 4) == 0xa000) /* rts */
    {
      if (testcond(opcode&0xf))
	{	      
	  pc_pop();
	  return ++curdsp.sequencer.pc;
	}
    }
  return 0;
}

/*
 * Type 21: Modify Address Register
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  0  0  0  1  0  0  1  0  0  0  0  0  0  0  0  0  0  0 -G --I-- --M--
 *
 */
unsigned int op_modify(unsigned int opcode)
{
  fprintf(stderr, "op_modify: unimplmeneted\n");
  return 0;
}

/*
 * Type 22: Conditional Trap (aka, Reserved)
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  0  0  0  1  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0 ----COND---
 *
 */
unsigned int op_trap(unsigned int opcode)
{
  fprintf(stderr, "op_trap: unimplmeneted\n");
  return 0;
}

/*
 * Type 23: DIVQ
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  0  0  0  0  1  1  1  0  0  0  0  0 ---XOP--  0  0  0  0  0  0  0  0
 *
 */
unsigned int op_divq(unsigned int opcode)
{
  fprintf(stderr, "op_divq: unimplmeneted\n");
  return 0;
}

/*
 * Type 24: DIVS
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  0  0  0  0  1  1  0  0  0  0 -YOP- -XOP-  0  0  0  0  0  0  0  0  0
 *
 */
unsigned int op_divs(unsigned int opcode)
{
  fprintf(stderr, "op_divs: unimplmeneted\n");
  return 0;
}

/*
 * Type 25: Saturate MR
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  0  0  0  0  1  0  1  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
 *
 */
unsigned int op_satmr(unsigned int opcode)
{
  fprintf(stderr, "op_satmr: unimplmeneted\n");
  return 0;
}

/*
 * Type 26: Stack Control
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  0  0  0  0  1  0  0  0  0  0  0  0  0  0  0  0  0  0 PP LP CP -SPP-
 *
 */
unsigned int op_stackctl(unsigned int opcode)
{
  fprintf(stderr, "op_stackctl: unimplmeneted\n");
  return 0;
}

/*
 * Type 27: Call/Jump on Flag In
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  0  0  0  0  0  1  1 ------------Address1--------------- -Ad2- -F -S
 *
 *  Address1 - 12LSBs
 *  Address2 -  2MSBs
 *         F - FIC
 *
 */
unsigned int op_jumpfi(unsigned int opcode)
{
  fprintf(stderr, "op_jumpfi: unimplmeneted\n");
  return 0;
}

/*
 * Type 28: Modify Flag Out
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  0  0  0  0  0  1  0  0  0  0  0 --F2- --F1- --F0- --FO- ---COND----
 *  
 *  F2 - FL2 
 *  F1 - FL1
 *  F0 - FL0
 *  FO - FLAG_OUT
 *
 */
unsigned int op_modifyfl(unsigned int opcode)
{
  fprintf(stderr, "op_modifyfl: unimplmeneted\n");
  return 0;
}

/*
 * Type 29: IO Memory Space R/W
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  0  0  0  0  0  0  1 -D -----------ADDR----------------- ---DREG----
 *
 */
unsigned int op_iorw(unsigned int opcode)
{
  fprintf(stderr, "op_iorw: unimplmeneted\n");
  return 0;
}

/*
 * Type 30: NOP
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
 *
 */
unsigned int op_nop(unsigned int opcode)
{
  return 0;
}

/*
 * Type 31: Idle
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  0  0  0  0  0  1  0  1  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
 *
 */
unsigned int op_idle(unsigned int opcode)
{
  /*
   * idle waits indefinitly for an interrupt
   */
  while (irq_checkdone()<0)
    {
      idler(); 
    }
  return 0;
}

/*
 * Type 32: Idle(n) [Slow Idle]
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  0  0  0  0  0  1  0  1  0  0  0  0  0  0  0  0  0  0  0 ---DV------
 *
 */
unsigned int op_idlen(unsigned int opcode)
{
  /*
   * idle(n) waits n cycles or for an interrupt, whichever is first 
   *
   * FIXME: above theory is wrong.  Hey mid, read the manuals before
   * making statements like that.
   */
  int n = opcode & 0xf;
  while (n  && (irq_checkdone()<0))
    {
      idler();
      n--;
    }
  return 0;
}
